#!/usr/bin/env python3

import argparse
import datetime
import json
import os
import time
import litellm
from litellm import EmbeddingResponse
from tenacity import (
    retry,
    retry_if_exception_type,
    stop_after_attempt,
    wait_random_exponential,
)
from tqdm import tqdm
from dotenv import load_dotenv
import sys
from IPython.core import ultratb
import warnings
import vdf_io
from vdf_io.constants import ID_COLUMN
from vdf_io.meta_types import NamespaceMeta, VDFMeta

from vdf_io.util import (
    get_final_data_path,
    get_parquet_files,
    read_parquet_progress,
    set_arg_from_input,
    set_arg_from_password,
)

litellm.suppress_debug_info = True
warnings.filterwarnings("ignore", module="litellm")
sys.excepthook = ultratb.FormattedTB(color_scheme="Linux", call_pdb=False)

load_dotenv()

def main():
    start_time = time.time()
    reembed()
    end_time = time.time()
    print(f"Time taken: {end_time - start_time:.2f} seconds")

def reembed():
    parser = argparse.ArgumentParser(description="Reembed a vector dataset")

    add_arguments_to_parser(parser)

    args = parser.parse_args()
    args = vars(args)

    if args.get("env_file_path"):
        load_dotenv(args.get("env_file_path"))

    take_input_from_cli_prompt(args)

    reembed_count = 0
    # open VDF_META.json
    handle_new_dataset(args)
    reembed_impl(args, reembed_count)

def reembed_impl(args, reembed_count):
    with open(os.path.join(args["dir"], "VDF_META.json"), "r+") as f:
        vdf_meta = json.load(f)
        for _, index_meta in tqdm(
            vdf_meta["indexes"].items(), desc="Iterating over indexes"
        ):
            new_vector_column = (
                f"vec_{args['text_column']}_{args['new_model_name'].replace('/', '_')}"
            )
            overwrite_bool = args["overwrite"]
            for namespace_meta in tqdm(index_meta, desc="Iterating over namespaces"):
                data_path = namespace_meta["data_path"]
                final_data_path = get_final_data_path(
                    os.getcwd(), args["dir"], data_path, args={}
                )
                parquet_files = get_parquet_files(
                    final_data_path,
                    args={},
                    id_column=vdf_meta.get("id_column", ID_COLUMN),
                )
                for file in tqdm(parquet_files, desc="Iterating over parquet files"):
                    file_path = os.path.join(final_data_path, file)
                    if (
                        new_vector_column in namespace_meta["vector_columns"]
                        and not overwrite_bool
                    ):
                        # ask user if they want to overwrite (y/n)
                        overwrite = args["overwrite"] or input(
                            f"{new_vector_column} already exists in vector_columns. Overwrite? (y/n): "
                        )
                        overwrite_bool = overwrite.lower() == "y"
                        if not overwrite_bool:
                            tqdm.write(
                                f"Skipping {file_path} because {new_vector_column} already exists in vector_columns.\n"
                                "Aborting reembedding."
                            )
                            exit()
                    if "dimensions" in args and args["dimensions"] is not None:
                        new_vector_column += f"_{args['dimensions']}"
                    tqdm.write(f"Reembedding {file_path}")
                    # read parquet file
                    df = read_parquet_progress(file_path, ID_COLUMN)
                    # get text column
                    text_column = args["text_column"]
                    if text_column not in df.columns:
                        text_columns = text_column.split("|")
                        if all([col in df.columns for col in text_columns]):
                            df[text_column] = df[text_columns].apply(
                                lambda x: " ".join(x.dropna().astype(str)), axis=1
                            )
                        else:
                            raise Exception(
                                f"Text column {text_column} not found in {file_path}"
                            )
                    # get embeddings
                    BATCH_SIZE = args.get("batch_size")
                    all_embeddings = []
                    for i in tqdm(
                        range(0, len(df), BATCH_SIZE), desc="Iterating over batches"
                    ):
                        batch_text = df[text_column][i : i + BATCH_SIZE].tolist()

                        embeddings = call_litellm(args, batch_text)
                        # add embeddings to df
                        # embeddings.data is a list of dicts. Each dict has keys "embedding" and "index"
                        # first sort by "index" and then extract "embedding"
                        vectors = [
                            x["embedding"]
                            for x in sorted(embeddings.data, key=lambda x: x["index"])
                        ]
                        dim = len(vectors[0])
                        all_embeddings.extend(vectors)
                    # add only the new_vector_column in parquet file
                    df[new_vector_column] = all_embeddings
                    if ID_COLUMN not in df.columns:
                        df[ID_COLUMN] = df.index
                    df.to_parquet(file_path)
                    tqdm.write(
                        f"Computed {len(all_embeddings)} vectors for {len(df)} rows in column:{new_vector_column} of {file_path}"
                    )
                    reembed_count += len(all_embeddings)
                # prepend new_vector_column to vector_columns in namespace_meta
                if new_vector_column not in namespace_meta["vector_columns"]:
                    namespace_meta["vector_columns"].insert(0, new_vector_column)
                else:
                    tqdm.write(
                        f"Warning: {new_vector_column} already exists in vector_columns. Overwriting."
                    )
                if "model_map" not in namespace_meta:
                    namespace_meta["model_map"] = {}
                    for vector_column in namespace_meta["vector_columns"]:
                        namespace_meta["model_map"][vector_column] = {
                            "model_name": namespace_meta["model_name"],
                            "text_column": args["text_column"],
                            "dimensions": namespace_meta["dimensions"],
                            "vector_column": vector_column,
                        }
                namespace_meta["model_map"][new_vector_column] = {
                    "model_name": args["new_model_name"],
                    "text_column": args["text_column"],
                    "dimensions": dim,
                    "vector_column": new_vector_column,
                }
                namespace_meta["model_name"] = args["new_model_name"]
                namespace_meta["dimensions"] = dim
        # write vdf_meta to VDF_META.json
        f.seek(0)
        json.dump(vdf_meta, f, indent=4)
        tqdm.write(
            f"Reembedding complete. Computed {reembed_count} vectors. Updated VDF_META.json"
        )

def handle_new_dataset(args):
    if not os.path.exists(os.path.join(args["dir"], "VDF_META.json")):
        # create VDF_META.json
        with open(os.path.join(args["dir"], "VDF_META.json"), "w") as f:
            # index_name is folder name of dir
            index_name = os.path.basename(args["dir"])
            # find parquet files in dir recursively
            import glob
            parquet_files = glob.glob(os.path.join(args["dir"], "**", "*.parquet"), recursive=True)
            if len(parquet_files) == 0:
                raise Exception("No parquet files found in the specified directory")
            total_vector_count = 0
            for file in parquet_files:
                # read schema of parquet file using pyarrow and get count of rows
                import pyarrow.parquet as pq
                table = pq.read_table(file)
                row_count = table.num_rows
                total_vector_count += row_count
            vdf_meta = VDFMeta(
                    version=vdf_io.__version__,
                    file_structure=[],
                    author=os.environ.get("USER"),
                    exported_from="reembed",
                    exported_at=datetime.datetime.now().astimezone().isoformat(),
                    id_column=ID_COLUMN,
                    indexes={
                        index_name: [
                            NamespaceMeta(
                                namespace="",
                                index_name=index_name,
                                total_vector_count=total_vector_count,
                                exported_vector_count=total_vector_count,
                                dimensions=-1,
                                model_name=None,
                                vector_columns=[],
                                data_path=".",
                                metric=None,
                            )
                        ]
                    },
                ).model_dump()
            json.dump(vdf_meta, f, indent=4)

def take_input_from_cli_prompt(args):
    set_arg_from_input(
        args,
        "new_model_name",
        "Enter the name of the new model to be used (default: 'text-embedding-3-small'): ",
        str,
        "text-embedding-3-small",
    )

    if args["new_model_name"].startswith("text-embedding-3"):
        set_arg_from_input(
            args,
            "dimensions",
            "Enter the dimensions of the new model to be used (default: 1536 for small, 3072 for large): ",
            int,
        )

    set_arg_from_password(
        args,
        "OPENAI_API_KEY",
        "Enter the OpenAI API key (hit return to skip): ",
        "OPENAI_API_KEY",
    )

    set_arg_from_input(
        args,
        "text_column",
        "Enter the name of the column containing text to be embedded (default: 'text'): ",
        str,
        "text",
    )

    set_arg_from_input(
        args,
        "dir",
        "Enter the directory of vector dataset in the VDF format: ",
        str,
    )

def add_arguments_to_parser(parser):
    parser.add_argument(
        "-d",
        "--dir",
        type=str,
        help="Directory of vector dataset in the VDF format",
        required=True,
    )

    parser.add_argument(
        "-m",
        "--new_model_name",
        type=str,
        help="Name of new model to be used",
    )

    parser.add_argument(
        "-t",
        "--text_column",
        type=str,
        help="Name of the column containing text to be embedded",
        default="text",
    )

    parser.add_argument(
        "--dimensions",
        type=int,
        help="Dimensions of the new model to be used",
    )

    parser.add_argument(
        "--env_file_path",
        type=str,
        help="Path to the .env file",
    )
    
    parser.add_argument(
        "--batch_size",
        type=int,
        help="Batch size for reembedding",
        default=100,
    )
    
    parser.add_argument(
        "--overwrite",
        type=bool,
        help="Overwrite existing vector columns",
        default=False,
        action=argparse.BooleanOptionalAction,
    )
use_sentence_transformers = False

@retry(
    wait=wait_random_exponential(multiplier=1, max=10),
    stop=stop_after_attempt(5),
    retry=retry_if_exception_type(Exception),
)
def call_litellm(args, batch_text):
    if args["new_model_name"].startswith("huggingface/"):
        return call_sentence_transformers(args, batch_text)
    global use_sentence_transformers
    if use_sentence_transformers:
        return call_sentence_transformers(args, batch_text)
    try:
        embeddings = litellm.embedding(
            model=args["new_model_name"],
            input=batch_text,
            dimensions=args.get("dimensions"),
        )
    except Exception as e:
        if e.message.startswith("Huggingface"):
            use_sentence_transformers = True
            embeddings = call_sentence_transformers(args, batch_text)
        else:
            raise e

    return embeddings


model = None


def call_sentence_transformers(args, batch_text):
    from sentence_transformers import SentenceTransformer
    # check global model
    global model
    if model is None:
        model = SentenceTransformer(
            args["new_model_name"].replace("huggingface/", ""), trust_remote_code=True
        )
    embeddings = model.encode(batch_text, show_progress_bar=True)
    return EmbeddingResponse(
        data=[
            {"index": i, "embedding": emb.tolist()} for i, emb in enumerate(embeddings)
        ]
    )


if __name__ == "__main__":
    main()



import json
from dotenv import load_dotenv
import numpy as np
from tqdm import tqdm
from grpc import RpcError
from typing import Any, Dict, List
from PIL import Image


import concurrent.futures

from qdrant_client import QdrantClient
from qdrant_client.http.exceptions import UnexpectedResponse
from qdrant_client.http.models import VectorParams, Distance, PointStruct

from vdf_io.names import DBNames
from vdf_io.util import (
    expand_shorthand_path,
    get_qdrant_id_from_id,
    set_arg_from_input,
    set_arg_from_password,
)
from vdf_io.import_vdf.vdf_import_cls import ImportVDB
from vdf_io.meta_types import NamespaceMeta

load_dotenv()


class ImportQdrant(ImportVDB):
    DB_NAME_SLUG = DBNames.QDRANT

    @classmethod
    def import_vdb(cls, args):
        """
        Import data to Qdrant
        """
        set_arg_from_input(
            args,
            "url",
            "Enter the URL of Qdrant instance (default: 'http://localhost:6334'): ",
            str,
            "http://localhost:6334",
        )
        set_arg_from_input(
            args,
            "prefer_grpc",
            "Whether to use GRPC. Recommended. (default: True): ",
            bool,
            True,
        )
        set_arg_from_input(
            args,
            "qdrant_local_persist_path",
            "Enter the path to the local persist directory (default: None): ",
            str,
            "DO_NOT_PROMPT",
        )
        set_arg_from_input(
            args,
            "parallel",
            "Enter the number of parallel processes of upload (default: 64): ",
            int,
            1,
        )
        set_arg_from_input(
            args,
            "max_retries",
            "Enter the maximum number of retries in case of a failure (default: 3): ",
            int,
            3,
        )
        set_arg_from_password(
            args, "qdrant_api_key", "Enter your Qdrant API key: ", "QDRANT_API_KEY"
        )
        qdrant_import = ImportQdrant(args)
        qdrant_import.upsert_data()
        return qdrant_import

    @classmethod
    def make_parser(cls, subparsers):
        parser_qdrant = subparsers.add_parser(
            DBNames.QDRANT, help="Import data to Qdrant"
        )
        parser_qdrant.add_argument(
            "-u",
            "--url",
            type=str,
            help="Qdrant instance url",
            default="http://localhost:6334",
        )
        parser_qdrant.add_argument(
            "--prefer_grpc",
            type=bool,
            help="Whether to use Qdrant's GRPC interface",
            default=True,
        )
        parser_qdrant.add_argument(
            "--qdrant_local_persist_path",
            type=str,
            help="Path to the local persist directory (default: None)",
            default=None,
        )
        parser_qdrant.add_argument(
            "--parallel",
            type=int,
            help="Number of parallel processes of upload (default: 5).",
            default=5,
        )
        parser_qdrant.add_argument(
            "--max_retries",
            type=int,
            help="Maximum number of retries in case of a failure (default: 3).",
            default=3,
        )
        parser_qdrant.add_argument(
            "--shard_key_selector",
            type=Any,
            help="Shard to be queried (default: None)",
            default=None,
        )

    def __init__(self, args):
        # call super class constructor
        super().__init__(args)
        url, api_key, prefer_grpc, path = (
            self.args.get("url", None),
            self.args.get("qdrant_api_key", None),
            self.args.get("prefer_grpc", True),
            expand_shorthand_path(self.args.get("qdrant_local_persist_path", None)),
        )
        if path:
            url = None
        if url:
            path = None
        self.client = QdrantClient(
            url=url,
            api_key=api_key,
            prefer_grpc=prefer_grpc,
            path=path,
        )

    def upsert_data(self):
        max_hit = False
        total_imported_count = 0
        # we know that the self.vdf_meta["indexes"] is a list
        index_meta: Dict[str, List[NamespaceMeta]] = {}
        for index_name, index_meta in tqdm(
            self.vdf_meta["indexes"].items(), desc="Importing indexes"
        ):
            tqdm.write(f"Importing data for index '{index_name}'")
            for namespace_meta in tqdm(index_meta, desc="Importing namespaces"):
                self.set_dims(namespace_meta, index_name)
                data_path = namespace_meta["data_path"]
                final_data_path = self.get_final_data_path(data_path)
                # list indexes
                collections = [
                    x.name for x in self.client.get_collections().collections
                ]
                # check if index exists
                new_collection_name = index_name + (
                    f"_{namespace_meta['namespace']}"
                    if namespace_meta["namespace"]
                    else ""
                )
                if new_collection_name not in collections:
                    # create index
                    try:

                        def get_nested_config(config, keys, default=None):
                            """Helper function to get nested dictionary values."""
                            if not config:
                                return default
                            for key in keys:
                                if not config:
                                    return default
                                config = config.get(key, {}) or {}
                            return config or default

                        index_config = namespace_meta.get("index_config", {})
                        dims = (
                            namespace_meta["dimensions"]
                            if "dimensions" in namespace_meta
                            else get_nested_config(
                                namespace_meta,
                                ["index_config", "params", "vectors"],
                                {},
                            ).get("size")
                        )
                        on_disk = get_nested_config(
                            namespace_meta,
                            ["index_config", "params", "vectors", "on_disk"],
                            None,
                        )
                        configs = [
                            "hnsw_config",
                            "optimizers_config",
                            "wal_config",
                            "quantization_config",
                            "on_disk_payload",
                            "sparse_vectors_config",
                        ]
                        (
                            hnsw_config,
                            optimizers_config,
                            wal_config,
                            quantization_config,
                            on_disk_payload,
                            sparse_vectors_config,
                        ) = [
                            get_nested_config(index_config, [config], None)
                            for config in configs
                        ]
                        distance = namespace_meta.get("metric", Distance.COSINE) or Distance.COSINE

                        self.client.create_collection(
                            collection_name=new_collection_name,
                            vectors_config=VectorParams(
                                size=dims, distance=distance, on_disk=on_disk
                            ),
                            sparse_vectors_config=sparse_vectors_config,
                            hnsw_config=hnsw_config,
                            optimizers_config=optimizers_config,
                            wal_config=wal_config,
                            quantization_config=quantization_config,
                            on_disk_payload=on_disk_payload,
                        )

                    except Exception as e:
                        tqdm.write(
                            f"Failed to create index '{new_collection_name}' {e}"
                        )
                        return
                prev_vector_count = self.client.get_collection(
                    collection_name=new_collection_name
                ).vectors_count
                if prev_vector_count > 0:
                    tqdm.write(
                        f"Index '{new_collection_name}' has {prev_vector_count} vectors before import"
                    )
                # Load the data from the parquet files
                parquet_files = self.get_parquet_files(final_data_path)

                vectors = {}
                metadata = {}
                vector_column_names, vector_column_name = self.get_vector_column_name(
                    new_collection_name, namespace_meta
                )
                for file in tqdm(parquet_files, desc="Iterating parquet files"):
                    file_path = self.get_file_path(final_data_path, file)
                    df = self.read_parquet_progress(file_path)
                    self.update_vectors(vectors, vector_column_name, df)
                    self.update_metadata(metadata, vector_column_names, df)
                    self.make_metadata_qdrant_compliant(metadata)
                    points = [
                        PointStruct(
                            id=get_qdrant_id_from_id(idx),
                            vector=vectors[idx],
                            payload=metadata.get(idx, {}),
                        )
                        for idx in vectors.keys()
                    ]

                    if total_imported_count + len(points) >= self.args["max_num_rows"]:
                        max_hit = True
                        points = points[
                            : self.args["max_num_rows"] - total_imported_count
                        ]
                        tqdm.write("Truncating data to limit to max rows")
                    try:
                        BATCH_SIZE = self.args.get("batch_size", 64) or 64
                        batches = list(divide_into_batches(points, BATCH_SIZE))
                        total_points = len(points)

                        num_parallel_threads = self.args.get("parallel", 5) or 5
                        with concurrent.futures.ThreadPoolExecutor(
                            max_workers=num_parallel_threads
                        ) as executor, tqdm(
                            total=total_points,
                            desc=f"Uploading points in batches of {BATCH_SIZE} in {num_parallel_threads} threads",
                        ) as pbar:
                            # Create a future to batch mapping to update progress bar correctly after each batch completion
                            future_to_batch = {
                                executor.submit(
                                    self.upsert_batch, batch, new_collection_name
                                ): batch
                                for batch in batches
                            }

                            for future in concurrent.futures.as_completed(
                                future_to_batch
                            ):
                                batch = future_to_batch[future]
                                try:
                                    # Attempt to get the result, which will re-raise any exceptions
                                    future.result()
                                    # Update the progress bar by the size of the successfully processed batch
                                    pbar.update(len(batch))
                                except Exception as e:
                                    tqdm.write(f"Batch upsert failed with error: {e}")
                                    # Optionally, you might want to handle failed batches differently
                        total_imported_count += len(points)
                        if total_imported_count >= self.args["max_num_rows"]:
                            max_hit = True
                    except (UnexpectedResponse, RpcError, ValueError) as e:
                        tqdm.write(
                            f"Failed to upsert data for collection '{new_collection_name}', {e}"
                        )
                        continue
                    vector_count = self.client.get_collection(
                        collection_name=new_collection_name
                    ).vectors_count
                    if max_hit:
                        break
                    # END parquet file loop
                tqdm.write(
                    f"Index '{new_collection_name}' has {vector_count} vectors after import"
                )
                tqdm.write(f"{vector_count - prev_vector_count} vectors were imported")
                if max_hit:
                    break
                # END namespace loop
            if max_hit:
                tqdm.write(
                    f"Max rows to be imported {self.args['max_num_rows']} hit. Exiting"
                )
                break
            # END index loop
        tqdm.write("Data import completed successfully.")
        self.args["imported_count"] = total_imported_count

    def make_metadata_qdrant_compliant(self, metadata):
        deleted_images = False
        parsed_json = False
        for k, v in metadata.items():
            deleted_images, parsed_json, zeroed_nan = self.normalize_dict(
                metadata, k, v
            )
        if deleted_images:
            tqdm.write("Images were deleted from metadata")
        if parsed_json:
            tqdm.write("Metadata was parsed to JSON")
        if zeroed_nan:
            tqdm.write("NaN values were replaced with 0 in metadata")

    def replace_nan_with_zero(self, data, zeroed_nan=False):
        if isinstance(data, dict):
            ret_val = {k: self.replace_nan_with_zero(v) for k, v in data.items()}
            for _, v in ret_val.items():
                if v[1]:
                    zeroed_nan = True
            return {k: v[0] for k, v in ret_val.items()}, zeroed_nan
        elif isinstance(data, list):
            ret_val = [self.replace_nan_with_zero(item) for item in data]
            return [x[0] for x in ret_val], any(x[1] for x in ret_val)
        elif isinstance(data, float) and np.isnan(data):
            return 0, True
        else:
            return data, False

    def normalize_dict(self, metadata, k, v):
        deleted_images = False
        parsed_json = False
        zeroed_nan = False
        # Check for np.nan and convert to 0 for scalar values
        if np.isscalar(v) and (
            (isinstance(v, (float, int)) and np.isnan(v))
            or (isinstance(v, str) and v.lower() == "nan")
        ):
            metadata[k] = 0
            zeroed_nan = True
        elif isinstance(v, np.ndarray):
            metadata[k] = v.tolist()
        elif isinstance(v, Image.Image):
            del metadata[k]
            deleted_images = True
        elif isinstance(v, bytes) or isinstance(v, str):
            if isinstance(v, bytes):
                metadata[k] = v.decode("utf-8")
            try:
                parsed_value = json.loads(metadata[k])
                # Replace nan with 0 in the parsed JSON object
                metadata[k], zeroed_nan_rec = self.replace_nan_with_zero(parsed_value)
                if zeroed_nan_rec:
                    zeroed_nan = True
                parsed_json = True
            except json.JSONDecodeError:
                pass
        elif isinstance(v, dict):
            for k2, v2 in v.items():
                (
                    deleted_images_rec,
                    parsed_json_rec,
                    zeroed_nan_rec,
                ) = self.normalize_dict(v, k2, v2)
                if zeroed_nan_rec:
                    zeroed_nan = True
                if deleted_images_rec:
                    deleted_images = True
                if parsed_json_rec:
                    parsed_json = True
        return deleted_images, parsed_json, zeroed_nan

    def upsert_batch(self, batch, new_collection_name):
        RETRIES = self.args.get("max_retries", 3)
        for attempt in range(RETRIES):
            try:
                self.client.upsert(
                    collection_name=new_collection_name,
                    points=batch,
                    shard_key_selector=self.args.get("shard_key_selector", None),
                    wait=True,
                )
                break  # Break the loop on success
            except Exception:
                if attempt == RETRIES - 1:
                    raise  # Re-raise the last exception if all retries fail
                else:
                    continue
        return len(batch)


# Function to divide your points into batches
def divide_into_batches(points, batch_size):
    for i in range(0, len(points), batch_size):
        yield points[i : i + batch_size]



#!/usr/bin/env python3

import os
import pandas as pd

from vdf_io.util import expand_shorthand_path, read_parquet_progress
from vdf_io.constants import ID_COLUMN

# script to get list of ids from directory of parquet files


def get_ids_from_parquet(directory):
    # for os.walk
    ids = set()
    for root, dirs, files in os.walk(expand_shorthand_path(dir)):
        # for each file in the directory
        for file in files:
            # if the file ends with .parquet
            if file.endswith(".parquet"):
                # get the path to the file
                path = os.path.join(root, file)
                # open the file
                # read only the id column
                try:
                    # Read only the 'id' column from the parquet file
                    df = read_parquet_progress(path, ID_COLUMN, columns=[ID_COLUMN])
                    ids.update(df[ID_COLUMN].tolist())
                except Exception as e:
                    print(f"Error reading {file}: {e}")
    return ids


dir = input("Enter directory of parquet files: ")
ids = get_ids_from_parquet(dir)
print(f"Found {len(ids)} ids")
print("Writing ids to ids.csv")
pd.DataFrame(list(ids)).sort_values(0).to_csv("ids.csv", index=False, header=False)

# if they are numeric ids, find the missing ids
if all([isinstance(i, int) for i in ids]):
    print("Finding missing ids")
    ids = set(ids)
    missing_ids = [i for i in range(1, max(ids)) if i not in ids]
    print(f"Found {len(missing_ids)} missing ids")
    print("Writing missing ids to missing_ids.csv")
    pd.DataFrame(missing_ids).to_csv("missing_ids.csv", index=False, header=False)
print("Done")



from pathlib import Path
from collections import OrderedDict
from getpass import getpass
import hashlib
import json
import os
from uuid import UUID
import pandas as pd
from io import StringIO
import sys
from tqdm import tqdm
from PIL import Image
from halo import Halo

from qdrant_client.http.models import Distance

from vdf_io.constants import ID_COLUMN
from vdf_io.names import DBNames


def sort_recursive(d):
    """
    Recursively sort the nested dictionary by its keys.
    """
    # if isinstance(d, list):
    #     return [sort_recursive(v) for v in d]
    # if isinstance(d, tuple):
    #     return tuple(sort_recursive(v) for v in d)
    # if isinstance(d, set):
    #     return list({sort_recursive(v) for v in d}).sort()
    if (
        isinstance(d, str)
        or isinstance(d, int)
        or isinstance(d, float)
        or isinstance(d, bool)
        or d is None
        or isinstance(d, OrderedDict)
    ):
        return d
    if hasattr(d, "attribute_map"):
        return sort_recursive(d.attribute_map)
    if not isinstance(d, dict):
        try:
            d = dict(d)
        except Exception:
            d = {"": str(d)}

    sorted_dict = OrderedDict()
    for key, value in sorted(d.items()):
        sorted_dict[key] = sort_recursive(value)

    return sorted_dict


def convert_to_consistent_value(d):
    """
    Convert a nested dictionary to a consistent string regardless of key order.
    """
    sorted_dict = sort_recursive(d)
    return json.dumps(sorted_dict, sort_keys=True)


def extract_data_hash(arg_dict_combined):
    arg_dict_combined_copy = arg_dict_combined.copy()
    data_hash = hashlib.md5(
        convert_to_consistent_value(arg_dict_combined_copy).encode("utf-8")
    )
    # make it 5 characters long
    data_hash = data_hash.hexdigest()[:5]
    return data_hash


def extract_numerical_hash(string_value):
    """
    Extract a numerical hash from a string
    """
    return int(hashlib.md5(string_value.encode("utf-8")).hexdigest(), 16)


def set_arg_from_input(args, arg_name, prompt, type_name=str, default_value=None):
    """
    Set the value of an argument from user input if it is not already present
    """
    if arg_name not in args or (
        args[arg_name] is None and default_value != "DO_NOT_PROMPT"
    ):
        inp = input(prompt)
        if inp == "":
            args[arg_name] = None if default_value is None else type_name(default_value)
        else:
            args[arg_name] = type_name(inp)
    return


def set_arg_from_password(args, arg_name, prompt, env_var_name):
    """
    Set the value of an argument from user input if it is not already present
    """
    if os.getenv(env_var_name) is not None:
        args[arg_name] = os.getenv(env_var_name)
    elif arg_name not in args or args[arg_name] is None:
        args[arg_name] = getpass(prompt)
    return


def expand_shorthand_path(shorthand_path):
    """
    Expand shorthand notations in a file path to a full path-like object.

    :param shorthand_path: A string representing the shorthand path.
    :return: A Path object representing the full path.
    """
    if shorthand_path is None:
        return None
    # Expand '~' to the user's home directory
    expanded_path = os.path.expanduser(shorthand_path)

    # Resolve '.' and '..' to get the absolute path
    full_path = Path(expanded_path).resolve()

    return str(full_path)


db_metric_to_standard_metric = {
    DBNames.PINECONE: {
        "cosine": Distance.COSINE,
        "euclidean": Distance.EUCLID,
        "dotproduct": Distance.DOT,
    },
    DBNames.QDRANT: {
        Distance.COSINE: Distance.COSINE,
        Distance.EUCLID: Distance.EUCLID,
        Distance.DOT: Distance.DOT,
        Distance.MANHATTAN: Distance.MANHATTAN,
    },
    DBNames.MILVUS: {
        "COSINE": Distance.COSINE,
        "IP": Distance.DOT,
        "L2": Distance.EUCLID,
    },
    DBNames.KDBAI: {
        "L2": Distance.EUCLID,
        "CS": Distance.COSINE,
        "IP": Distance.DOT,
    },
    DBNames.VERTEXAI: {
        "DOT_PRODUCT_DISTANCE": Distance.DOT,
        "SQUARED_L2_DISTANCE": Distance.EUCLID,
        "COSINE_DISTANCE": Distance.COSINE,
        "L1_DISTANCE": Distance.MANHATTAN,
    },
}


def standardize_metric(metric, db):
    """
    Standardize the metric name to the one used in the standard library.
    """
    if (
        db in db_metric_to_standard_metric
        and metric in db_metric_to_standard_metric[db]
    ):
        return db_metric_to_standard_metric[db][metric]
    else:
        raise Exception(f"Invalid metric '{metric}' for database '{db}'")


def standardize_metric_reverse(metric, db):
    """
    Standardize the metric name to the one used in the standard library.
    """
    if (
        db in db_metric_to_standard_metric
        and metric in db_metric_to_standard_metric[db].values()
    ):
        for key, value in db_metric_to_standard_metric[db].items():
            if value == metric:
                return key
    else:
        raise Exception(f"Invalid metric '{metric}' for database '{db}'")


def get_final_data_path(cwd, dir, data_path, args):
    if args.get("hf_dataset", None):
        return data_path
    final_data_path = os.path.join(cwd, dir, data_path)
    if not os.path.isdir(final_data_path):
        raise Exception(
            f"Invalid data path\n"
            f"data_path: {data_path},\n"
            f"Joined path: {final_data_path}\n"
            f"Current working directory: {cwd}\n"
            f"Command line arg (dir): {dir}"
        )
    return final_data_path


def list_configs_and_splits(name):
    if "HUGGING_FACE_TOKEN" not in os.environ:
        yield "train", None
    import requests

    headers = {"Authorization": f"Bearer {os.environ['HUGGING_FACE_TOKEN']}"}
    API_URL = f"https://datasets-server.huggingface.co/splits?dataset={name}"

    def query():
        response = requests.get(API_URL, headers=headers)
        return response.json()

    data = query()
    if "splits" in data:
        for split in data["splits"]:
            if "config" in split:
                yield split["split"], split["config"]
            else:
                yield split["split"], None
    else:
        yield "train", None


def get_parquet_files(data_path, args, temp_file_paths=[], id_column=ID_COLUMN):
    # Load the data from the parquet files
    if args.get("hf_dataset", None):
        if args.get("max_num_rows", None):
            from datasets import load_dataset

            total_rows_loaded = 0
            for i, (split, config) in enumerate(
                list_configs_and_splits(args.get("hf_dataset"))
            ):
                tqdm.write(f"Split: {split}, Config: {config}")
                ds = load_dataset(
                    args.get("hf_dataset"), name=config, split=split, streaming=True
                )
                with Halo(text="Taking a subset of the dataset", spinner="dots"):
                    it_ds = ds.take(args.get("max_num_rows") - total_rows_loaded)
                with Halo(text="Converting to pandas dataframe", spinner="dots"):
                    df = pd.DataFrame(it_ds)
                df = cleanup_df(df)
                if id_column not in df.columns:
                    # remove all rows
                    tqdm.write(
                        (
                            f"ID column '{id_column}' not found in parquet file '{data_path}'."
                            f" Skipping split '{split}', config '{config}'."
                        )
                    )
                    continue
                total_rows_loaded += len(df)
                temp_file_path = f"{os.getcwd()}/temp_{args['hash_value']}_{i}.parquet"
                with Halo(text="Saving to parquet", spinner="dots"):
                    df.to_parquet(temp_file_path)
                temp_file_paths.append(temp_file_path)
                if total_rows_loaded >= args.get("max_num_rows"):
                    break
            return temp_file_paths
        from huggingface_hub import HfFileSystem

        fs = HfFileSystem()
        return [
            "hf://" + x
            for x in fs.glob(
                f"datasets/{args.get('hf_dataset')}/{data_path if data_path!='.' else ''}/**.parquet"
            )
        ]
    if not os.path.isdir(data_path):
        if data_path.endswith(".parquet"):
            return [data_path]
        else:
            raise Exception(f"Invalid data path '{data_path}'")
    else:
        parquet_files = sorted(
            [file for file in os.listdir(data_path) if file.endswith(".parquet")]
        )
        return parquet_files


def cleanup_df(df):
    for col in df.columns:
        if df[col].dtype == "object":
            first_el = df[col].iloc[0]
            # if isinstance(first_el, bytes):
            #     df[col] = df[col].apply(lambda x: x.decode("utf-8"))
            if isinstance(first_el, Image.Image):
                # delete the image column
                df = df.drop(columns=[col])
                tqdm.write(
                    f"Warning: Image column '{col}' detected. Image columns are not supported in parquet files. The column has been removed."
                )

    return df


# Function to recursively print help messages
def print_help_recursively(parser, level=0):
    # Temporarily redirect stdout to capture the help message
    old_stdout = sys.stdout
    sys.stdout = StringIO()

    # Print the current parser's help message
    parser.print_help()

    # Retrieve and print the help message from the StringIO object
    help_message = sys.stdout.getvalue()
    sys.stdout = old_stdout  # Restore stdout

    # Print the captured help message with indentation for readability
    print("\n" + "\t" * level + "Help message for level " + str(level) + ":")
    for line in help_message.split("\n"):
        print("\t" * level + line)

    # Check if the current parser has subparsers
    if hasattr(parser, "_subparsers"):
        for _, subparser in parser._subparsers._group_actions[0].choices.items():
            # Recursively print help for each subparser
            print_help_recursively(subparser, level + 1)


def is_str_uuid(id_str):
    try:
        uuid_obj = UUID(id_str)
        return str(uuid_obj)
    except ValueError:
        return False


def get_qdrant_id_from_id(idx):
    if isinstance(idx, int) or idx.isdigit():
        return int(idx)
    elif not is_str_uuid(idx):
        hex_string = hashlib.md5(idx.encode("UTF-8")).hexdigest()
        return str(UUID(hex=hex_string))
    else:
        return str(UUID(idx))


def read_parquet_progress(file_path, id_column, **kwargs):
    if file_path.startswith("hf://"):
        from huggingface_hub import HfFileSystem
        from huggingface_hub import hf_hub_download

        fs = HfFileSystem()
        resolved_path = fs.resolve_path(file_path)
        cache_path = hf_hub_download(
            repo_id=resolved_path.repo_id,
            filename=resolved_path.path_in_repo,
            repo_type=resolved_path.repo_type,
        )
        file_path_to_be_read = cache_path
    else:
        file_path_to_be_read = file_path
    # read schema of the parquet file to check if columns are present
    from pyarrow import parquet as pq

    schema = pq.read_schema(file_path_to_be_read)
    # list columns
    columns = schema.names
    # if kwargs has columns, check if all columns are present
    cols = set()
    cols.add(id_column)
    return_empty = False
    if "columns" in kwargs:
        for col in kwargs["columns"]:
            cols.add(col)
            if col not in columns:
                tqdm.write(
                    f"Column '{col}' not found in parquet file '{file_path_to_be_read}'. Returning empty DataFrame."
                )
                return_empty = True
    if return_empty:
        return pd.DataFrame(columns=list(cols))
    df = pd.read_parquet(file_path_to_be_read, **kwargs)
    return df



#!/usr/bin/env python3

import argparse
import re


def bump_version(version_file, part="patch"):
    # Read the current version
    with open(version_file, "r") as file:
        content = file.read()

    # Find the version number using regex
    version_match = re.search(r"version\s*=\s*['\"]([^'\"]+)['\"]", content)
    if version_match:
        version = version_match.group(1)
    else:
        raise ValueError("Version number not found in setup.py.")

    # Split version into components
    major, minor, patch = map(int, version.split("."))

    # Increment the appropriate part of the version
    if part == "major":
        major += 1
        minor = 0  # Reset minor and patch numbers when major is incremented
        patch = 0
    elif part == "minor":
        minor += 1
        patch = 0  # Reset patch number when minor is incremented
    elif part == "patch":
        patch += 1
    else:
        raise ValueError("Invalid part specified. Use 'major', 'minor', or 'patch'.")

    # Combine the parts back into a version string
    new_version = f"{major}.{minor}.{patch}"

    # Replace the old version with the new version in the content
    new_content = re.sub(
        r"(version\s*=\s*['\"])[^'\"]+(['\"])", rf"\g<1>{new_version}\g<2>", content
    )

    # Write the new content back to the file
    with open(version_file, "w") as file:
        file.write(new_content)

    return new_version


def main():
    parser = argparse.ArgumentParser(description="Bump version number in setup.py.")
    parser.add_argument(
        "version_file",
        nargs="?",
        default="setup.py",
        help="Path to the version file (default: setup.py)",
    )
    parser.add_argument(
        "--part",
        choices=["major", "minor", "patch"],
        default="patch",
        help="Part of version to bump (default: patch)",
    )
    args = parser.parse_args()

    new_version = bump_version(args.version_file, args.part)

    # run `sed -i '' "s/__version__ = .*/__version__ = '$version_number'/" src/vdf_io/__init__.py`
    # Replace the version number in the file
    init_file = "src/vdf_io/__init__.py"
    content = f'__version__ = "{new_version}"\n'
    for line in open(init_file, "r"):
        if "__version__" in line:
            content = line.replace(line, content)
            break
    with open(init_file, "w") as file:
        file.write(content)

    print(f"Updated version: {new_version}")


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

import argparse
import os
import sys
import time
import traceback
from dotenv import load_dotenv
import warnings

# Suppress specific warnings
warnings.simplefilter("ignore", ResourceWarning)
warnings.filterwarnings("ignore", module="numpy")


import sentry_sdk  # noqa: E402
from opentelemetry import trace  # noqa: E402
from opentelemetry.propagate import set_global_textmap  # noqa: E402
from opentelemetry.sdk.trace import TracerProvider  # noqa: E402
from sentry_sdk.integrations.opentelemetry import (  # noqa: E402
    SentrySpanProcessor,  # noqa: E402
    SentryPropagator,  # noqa: E402
)  # noqa: E402

import vdf_io  # noqa: E402
from vdf_io.export_vdf.milvus_export import ExportMilvus  # noqa: E402
from vdf_io.export_vdf.pinecone_export import ExportPinecone  # noqa: E402
from vdf_io.export_vdf.qdrant_export import ExportQdrant  # noqa: E402
from vdf_io.export_vdf.kdbai_export import ExportKDBAI  # noqa: E402
from vdf_io.export_vdf.vertexai_vector_search_export import (  # noqa: E402
    ExportVertexAIVectorSearch,  # noqa: E402
)  # noqa: E402
from vdf_io.names import DBNames  # noqa: E402
from vdf_io.scripts.check_for_updates import check_for_updates  # noqa: E402
from vdf_io.scripts.push_to_hub_vdf import push_to_hub  # noqa: E402


load_dotenv()

DEFAULT_MAX_FILE_SIZE = 1024  # in MB


if os.environ.get("DISABLE_TELEMETRY_VECTORIO", False) != "1":
    sentry_sdk.init(
        dsn="https://4826b78415eeaf0135c12416e222596d@o1284436.ingest.sentry.io/4506716331573248",
        enable_tracing=True,
        # set the instrumenter to use OpenTelemetry instead of Sentry
        instrumenter="otel",
        default_integrations=False,
    )


provider = TracerProvider()
provider.add_span_processor(SentrySpanProcessor())
trace.set_tracer_provider(provider)
set_global_textmap(SentryPropagator())

tracer = trace.get_tracer(__name__)


def main():
    with tracer.start_as_current_span("export_vdf_cli_main") as span:
        try:
            run_export(span)
            sentry_sdk.flush()
        except Exception as e:
            sentry_sdk.flush()
            print(f"Error: {e}")
            traceback.print_exc()
            sys.exit(1)
        finally:
            sentry_sdk.flush()
    sentry_sdk.flush()
    return


ARGS_ALLOWLIST = [
    "vector_database",
    "library_version",
    "hash_value",
    "exported_count",
]

slug_to_export_func = {
    DBNames.PINECONE: ExportPinecone.export_vdb,
    DBNames.QDRANT: ExportQdrant.export_vdb,
    DBNames.KDBAI: ExportKDBAI.export_vdb,
    DBNames.MILVUS: ExportMilvus.export_vdb,
    DBNames.VERTEXAI: ExportVertexAIVectorSearch.export_vdb,
}

slug_to_parser_func = {
    DBNames.PINECONE: ExportPinecone.make_parser,
    DBNames.QDRANT: ExportQdrant.make_parser,
    DBNames.KDBAI: ExportKDBAI.make_parser,
    DBNames.MILVUS: ExportMilvus.make_parser,
    DBNames.VERTEXAI: ExportVertexAIVectorSearch.make_parser,
}


def run_export(span):
    parser = argparse.ArgumentParser(
        description="Export data from various vector databases to the VDF format for vector datasets"
    )
    make_common_options(parser)
    subparsers = parser.add_subparsers(
        title="Vector Databases",
        description="Choose the vectors database to export data from",
        dest="vector_database",
    )

    db_choices = slug_to_export_func.keys()
    for db in db_choices:
        slug_to_parser_func[db](subparsers)

    args = parser.parse_args()
    # convert args to dict
    args = vars(args)
    args["library_version"] = vdf_io.__version__

    t_start = time.time()
    if (
        ("vector_database" not in args)
        or (args["vector_database"] is None)
        or (args["vector_database"] not in db_choices)
    ):
        print("Please choose a vector database to export data from:", db_choices)
        return

    if args["vector_database"] in db_choices:
        export_obj = slug_to_export_func[args["vector_database"]](args)
    else:
        print("Invalid vector database")
        args["vector_database"] = input("Enter the name of vector database to export: ")
        sys.argv.extend(["--vector_database", args["vector_database"]])
        main()
    t_end = time.time()

    for key in list(export_obj.args.keys()):
        if key in ARGS_ALLOWLIST:
            span.set_attribute(key, export_obj.args[key])

    # formatted time
    print(f"Export to disk completed. Exported to: {export_obj.vdf_directory}/")
    print(
        "Time taken to export data: ",
        time.strftime("%H:%M:%S", time.gmtime(t_end - t_start)),
    )
    span.set_attribute("export_time", t_end - t_start)
    if args["push_to_hub"]:
        push_to_hub(export_obj, args)

    check_for_updates()


def make_common_options(parser):
    parser.add_argument(
        "-m",
        "--model_name",
        type=str,
        help="Name of model used",
        default="text-embedding-ada-002",
    )
    parser.add_argument(
        "--max_file_size",
        type=int,
        help="Maximum file size in MB (default: 1024)",
        default=DEFAULT_MAX_FILE_SIZE,
    )

    parser.add_argument(
        "--push_to_hub",
        type=bool,
        help="Push to hub",
        default=False,
        action=argparse.BooleanOptionalAction,
    )
    parser.add_argument(
        "--public",
        type=bool,
        help="Make dataset public (default: False)",
        default=False,
        action=argparse.BooleanOptionalAction,
    )


if __name__ == "__main__":
    main()



import requests
import pkg_resources as pkg

import vdf_io


def check_for_updates():
    package_name = "vdf-io"
    response = requests.get(f"https://pypi.org/pypi/{package_name}/json")
    if response.status_code == 200:
        latest_version = response.json()["info"]["version"]
        current_version = pkg.parse_version(vdf_io.__version__)
        latest_version = pkg.parse_version(latest_version)
        if current_version.release[:2] < latest_version.release[:2]:
            print(
                f"Current version: {vdf_io.__version__}. Update available: {latest_version}. Run `pip install --upgrade {package_name}` to update."
            )


def main():
    check_for_updates()


if __name__ == "__main__":
    main()



from dotenv import load_dotenv
from tqdm import tqdm
import json

from pymilvus import (
    connections,
    utility,
    Collection,
    CollectionSchema,
    FieldSchema,
    DataType,
)

from vdf_io.names import DBNames
from vdf_io.util import (
    set_arg_from_input,
    set_arg_from_password,
    standardize_metric_reverse,
)
from vdf_io.import_vdf.vdf_import_cls import ImportVDB


load_dotenv()


class ImportMilvus(ImportVDB):
    DB_NAME_SLUG = DBNames.MILVUS

    @classmethod
    def import_vdb(cls, args):
        """
        Import data to Milvus
        """
        set_arg_from_input(
            args,
            "uri",
            "Enter the Milvus URI (default: 'http://localhost:19530'): ",
            str,
            "http://localhost:19530",
        )
        set_arg_from_password(
            args,
            "token",
            "Enter your Milvus/Zilliz token (hit enter to skip): ",
            "ZILLIZ_CLOUD_TOKEN",
        )
        milvus_import = ImportMilvus(args)
        milvus_import.upsert_data()
        return milvus_import

    @classmethod
    def make_parser(cls, subparsers):
        parser_milvus = subparsers.add_parser(
            DBNames.MILVUS, help="Import data to Milvus"
        )
        parser_milvus.add_argument(
            "-u", "--uri", type=str, help="URI of Milvus instance"
        )
        parser_milvus.add_argument("-t", "--token", type=str, help="Milvus token")

    def __init__(self, args):
        # call super class constructor
        super().__init__(args)
        if args is None:
            args = {}
        assert isinstance(args, dict), "Invalid args."
        super().__init__(args)

        uri = self.args.get("uri", "http://localhost:19530")
        token = self.args.get("token", "")
        connections.connect(uri=uri, token=token)

    def upsert_data(self):
        max_hit = False
        total_imported_count = 0
        # we know that the self.vdf_meta["indexes"] is a list
        for collection_name, index_meta in self.vdf_meta["indexes"].items():
            # load data
            print(f'Importing data for collection "{collection_name}"')
            for namespace_meta in tqdm(index_meta, desc="Importing namespaces"):
                self.set_dims(namespace_meta, collection_name)
                data_path = namespace_meta["data_path"]
                index_name = collection_name + (
                    f'_{namespace_meta["namespace"]}'
                    if namespace_meta["namespace"]
                    else ""
                )
                _, vector_column_name = self.get_vector_column_name(
                    collection_name, namespace_meta
                )
                # replace - with _
                old_vector_column_name = vector_column_name
                vector_column_name = vector_column_name.replace("-", "_")
                index_name = index_name.replace("-", "_")
                print(f"Index name: {index_name}")

                # check if collection exists
                if utility.has_collection(index_name):
                    collection = Collection(index_name)
                    f_vector = None
                    f_pk = None
                    for f in collection.schema.fields:
                        if f.dtype.value in [100, 101]:
                            f_vector = f
                        if hasattr(f, "is_primary") and f.is_primary:
                            f_pk = f
                else:
                    # create collection
                    try:
                        f_pk = FieldSchema(
                            name=self.id_column,
                            dtype=DataType.VARCHAR,
                            max_length=65_535,
                            is_primary=True,
                            auto_id=False,
                        )
                        f_vector = FieldSchema(
                            name=vector_column_name,
                            dtype=DataType.FLOAT_VECTOR,
                            dim=namespace_meta["dimensions"],
                        )
                        schema = CollectionSchema(
                            fields=[f_pk, f_vector], enable_dynamic_field=True
                        )
                        collection = Collection(name=index_name, schema=schema)
                    except Exception as e:
                        print(f'Failed to create collection "{index_name}"', e)
                        raise RuntimeError("Failed to create collection.") from e

                # check if index exists
                if index_name in [index.index_name for index in collection.indexes]:
                    print(f"Using existed index: {index_name}")
                else:
                    # create index
                    try:
                        index_params = {
                            "metric_type": standardize_metric_reverse(
                                namespace_meta["metric"], self.DB_NAME_SLUG
                            ),
                            "index_type": "AUTOINDEX",
                        }
                        collection.create_index(
                            field_name=f_vector.name, index_params=index_params
                        )
                    except Exception as e:
                        print(
                            f"Faild to create index {index_name} for collection {index_name}."
                        )
                        raise RuntimeError("Failed to create index.") from e

                prev_vector_count = collection.num_entities
                if prev_vector_count > 0:
                    print(
                        f'Collection "{index_name}" has {prev_vector_count} vectors before import'
                    )

                # Load the data from the parquet files
                final_data_path = self.get_final_data_path(data_path)
                parquet_files = self.get_parquet_files(final_data_path)

                num_inserted = 0
                for file in tqdm(parquet_files, desc="Iterating parquet files"):
                    file_path = self.get_file_path(final_data_path, file)
                    df = self.read_parquet_progress(file_path)
                    df[self.id_column] = df[self.id_column].apply(lambda x: str(x))
                    data_rows = []

                    for _, row in df.iterrows():
                        row[old_vector_column_name] = self.extract_vector(
                            row[old_vector_column_name]
                        )
                        row = json.loads(row.to_json())
                        # replace old_vector_column_name with vector_column_name
                        if old_vector_column_name != vector_column_name:
                            row[vector_column_name] = row[old_vector_column_name]
                            del row[old_vector_column_name]
                        assert isinstance(row[f_pk.name], str), row[f_pk.name]
                        assert isinstance(row[f_vector.name][0], float), type(
                            row[f_vector.name][0]
                        )
                        data_rows.append(row)
                    BATCH_SIZE = self.args.get("batch_size", 1000) or 1000
                    current_batch_size = BATCH_SIZE
                    tqdm.write(
                        f"Inserting {len(data_rows)} rows in batches of {BATCH_SIZE}"
                    )
                    if total_imported_count + BATCH_SIZE >= self.args["max_num_rows"]:
                        data_rows = data_rows[
                            : self.args["max_num_rows"] - total_imported_count
                        ]
                        max_hit = True
                    i = 0
                    pbar = tqdm(total=len(data_rows), desc="Upserting data in batches")
                    while i < len(data_rows):
                        try:
                            mr = collection.upsert(
                                data_rows[i : i + current_batch_size]
                            )
                            num_inserted += mr.upsert_count
                            total_imported_count += mr.upsert_count
                            i += current_batch_size
                            pbar.update(current_batch_size)
                        except Exception as e:
                            tqdm.write(f"Error inserting data: {e}")
                            # reduce batch size
                            current_batch_size = current_batch_size * 2 // 3
                            tqdm.write(f"Reducing batch size to {current_batch_size}")
                            continue
                    if max_hit:
                        break
                collection.flush()
                vector_count = collection.num_entities
                print(f"Index '{index_name}' has {vector_count} vectors after import")
                print(f"{num_inserted} vectors were imported")
        print("Data import completed successfully.")
        self.args["imported_count"] = total_imported_count
